/**
  *  \file playvcr.cc
  *  \brief Standalone VCR Player
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <SDL.h>
#include "arch/platform.h"
#include "client/gdchooser.h"
#include "client/vcrplay.h"
#include "cpluslib/format.h"
#include "io/streams.h"
#include "game/config.h"
#include "game/global.h"
#include "game/spec.h"
#include "game/struct.h"
#include "game/vcrdb.h"
#include "gfx/init.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/widgets.h"
#include "util/cmdline.h"
#include "util/console.h"
#include "util/debug.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"
#include "util/recode.h"
#include "sound/sound.h"
#include "client/sound-conf.h"
#include "game/pref.h"
#include "util/confpars.h"
#include "cpluslib/holder.h"
#include "defs.h"
#include "gfx/blit.h"
#include "client/scr-sim.h"

// #define SIMULATOR_SUPPORT

#ifdef FLAK_SUPPORT
#  include "flak/flakdefs.h"
#  include "flak/flak.h"
#  include "flak/vis.h"
#  include "flak/dlg-select.h"
#endif

static bool optDirectory = false;

namespace {
    /** Option Parser for playvcr. */
    class PlayvcrOptionParser : public StandardCommandLineParser {
        enum { MAX_NON_OPTIONS = 2 };
        int num_nonoptions;
        string_t nonopts[MAX_NON_OPTIONS];
        int battle_nr;
        static void help(StandardCommandLineParser&, const Option&);
        static void setCharset(StandardCommandLineParser&, const Option&);
     public:
        PlayvcrOptionParser();
        ~PlayvcrOptionParser();
        void handleNonoption(const string_t& str);
        void handleError(const string_t& msg);
        int getNumNonoptions() const { return num_nonoptions; }
        int getBattleNumber() const { return battle_nr; }
        void parseFinished();

        void setFileName(const string_t& s) { nonopts[0] = s; }
        void setDirName(const string_t& s) { nonopts[1] = s; }

        string_t getFileName() const { return nonopts[0]; }
        string_t getDirName() const { return nonopts[1]; }

        static PlayvcrOptionParser Instance;
    };

    /** Console. Displays the console messages in the background. */
    class PlayvcrConsole : public UIFixedWidget {
        SignalConnection conn;
     public:
        PlayvcrConsole();
        ~PlayvcrConsole();
        void draw(GfxCanvas& can, UpdateInfo& ui);
        void update(const Console&, int);
    };

    class PlayvcrImageWidget : public UIWidget {
        Ptr<GfxPixmap> pix;
     public:
        PlayvcrImageWidget(Ptr<GfxPixmap> pix);
        void draw(GfxCanvas& can, UpdateInfo& ui);
        void getLayoutInfo(LayoutInfo& i);
    };

    class PlayvcrSimpleButton : public UIAbstractButton {
     public:
        PlayvcrSimpleButton(int id, const char* name, UIKey key);
        void draw(GfxCanvas& can, UpdateInfo& ui);
        void getLayoutInfo(LayoutInfo&);
    };
}

PlayvcrOptionParser PlayvcrOptionParser::Instance;

/** Initialize. */
PlayvcrOptionParser::PlayvcrOptionParser()
    : num_nonoptions(0), battle_nr(0)
{
    static const Option options[] = {
        { "help", help, 0, false,
          N_(" -help              Show help screen") },
        { "dir", &optSetValue<bool,true>, &optDirectory, false,
          N_(" -dir               Start with directory selection") },
        { "charset", setCharset, 0, true,
          N_(" -charset EXT[:INT] Set character set translation") }
    };
    addOptionSet(options, countof(options));
}

PlayvcrOptionParser::~PlayvcrOptionParser()
{ }

/** Handle Non-Options. The first non-option is the file name, the
    second (optional) is the directory. A battle number may also
    be specified. */
void
PlayvcrOptionParser::handleNonoption(const string_t& str)
{
    int tmp;
    if (! battle_nr && stringValue(str, tmp) && tmp > 0) {
        battle_nr = tmp;
    } else {
        if (num_nonoptions >= MAX_NON_OPTIONS)
            handleError(_("Too many parameters specified."));
        nonopts[num_nonoptions++] = str;
    }
}

/* FIXME: this very function also appears in main.cc */
void
PlayvcrOptionParser::setCharset(StandardCommandLineParser& par,
                                const StandardCommandLineParser::Option&)
{
    string_t cs = par.getArg();
    string_t intrn = "latin1";
    strSplit(cs, cs, intrn, ":");
    CharacterSet ics = Recoder::getCharsetByName(intrn);
    CharacterSet ecs = Recoder::getCharsetByName(cs);
    if (!ics.isValid())
        console.write(LOG_WARN, format(_("Character set `%s' is unknown"), intrn));
    if (!ecs.isValid())
        console.write(LOG_WARN, format(_("Character set `%s' is unknown"), cs));
    Recoder::setCharsets(ics, ecs);
}

/** Show help screen. */
void
PlayvcrOptionParser::help(StandardCommandLineParser& par,
                          const StandardCommandLineParser::Option&)
{
    console << _("Usage: playvcr [-options] vcrfile [gamedir] [battlenr]\n"
                 "\n"
                 "Valid options:\n");
    par.printHelp(console);
    console << _("\n(c) 2001-2004 Stefan Reuther <Streu@gmx.de>\n");
    Platform::showOutput();
    std::exit(0);
}

/** Error handler. Instead of throwing an exception, show a message
    and bail out. */
void
PlayvcrOptionParser::handleError(const string_t& str)
{
    console.write(LOG_ERROR, str);
    console.write(LOG_ERROR, _("Try `playvcr --help' for help."));
    Platform::showOutput();
    std::exit(1);
}

/** Parsing finished. Postprocess input. */
void
PlayvcrOptionParser::parseFinished()
{
    if (optDirectory) {
        if (getNumNonoptions() > 1)
            handleError(_("Only one parameter expected when `-dir' is specified."));
    } else {
        if (getNumNonoptions() == 1) {
            nonopts[1] = Directory::getDirName(nonopts[0]);
            console << format(_("Using game directory `%s'\n"), nonopts[1]);
        }
    }
}

/***************************** PlayvcrConsole ****************************/

/** Initialize. */
PlayvcrConsole::PlayvcrConsole()
    : UIFixedWidget(0), conn(console.console_changed.add(this, &PlayvcrConsole::update))
{ }

PlayvcrConsole::~PlayvcrConsole()
{ }

void
PlayvcrConsole::draw(GfxCanvas& can, UpdateInfo& ui)
{
    const int font = FONT_NORMAL;
    int nlines = extent.h / font_heights[font];
    GfxParameters parm;
    parm.bgcolor = standard_colors[COLOR_BLACK];
    int y = extent.h;
    for (int i = 0; i < nlines; ++i) {
        parm.color = standard_colors[COLOR_GRAYSCALE+15 - (i*10) / nlines];
        parm.font = fonts[font];
        y -= font_heights[font];
        outTextF(can, parm, extent.x,
                 extent.y + y, extent.w,
                 (i >= console.getScrollbackFill()
                  ? string_t()
                  : console.getLine(i).text));
    }
    if (y)
        drawSolidBar(can, GfxRect(extent.x, extent.y, extent.w, y),
                     standard_colors[COLOR_BLACK]);
    ui.update(extent);
}

void
PlayvcrConsole::update(const Console&, int)
{
    drawWidget(false);
}

PlayvcrImageWidget::PlayvcrImageWidget(Ptr<GfxPixmap> pix)
    : UIWidget(0), pix(pix)
{
    setExtent(GfxRect(pix->getSystemSurface()));
}

void
PlayvcrImageWidget::draw(GfxCanvas& can, UpdateInfo& ui)
{
    pix->blit(can, extent.x, extent.y);
    ui.update(extent);
}

void
PlayvcrImageWidget::getLayoutInfo(LayoutInfo& i)
{
    i.flags = li_NoLayout;
}

PlayvcrSimpleButton::PlayvcrSimpleButton(int id, const char* name, UIKey key)
    : UIAbstractButton(id, name, key, aNothing)
{ }

void
PlayvcrSimpleButton::draw(GfxCanvas& can, UpdateInfo& ui)
{
    GfxParameters parm;
    if (hasState(st_Selected))
        parm.color = standard_colors[COLOR_GRAYSCALE + 14];
    else
        parm.color = standard_colors[COLOR_GRAYSCALE + 12];
    parm.bgcolor = standard_colors[COLOR_BLACK];
    parm.font = fonts[FONT_NORMAL];
    outTextF(can, parm, extent.x, extent.y,
             extent.w, getName());
    ui.update(getExtent());
}

void
PlayvcrSimpleButton::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = GfxPoint(fonts[FONT_NORMAL]->getTextWidth(getName()),
                                          font_heights[FONT_NORMAL]);
}

/*************************** Directory Chooser ***************************/

namespace {
    /** Game chooser for VCR. Displays the VCR count for each race. */
    class WVcrGameChooser : public WGameRaceChooser {
        WGamedirChooser* gdc;
        string_t file_name[NUM_PLAYERS];
     protected:
        void handleMovement(int);
     public:
        WVcrGameChooser(int id, WGamedirChooser* gdc);
        void scan(Ptr<Directory> dir);
        string_t getFileName(int player) const
            { return file_name[player-1]; }
    };

    /** Game chooser window. Provides initialisation and some event
        dispatching. */
    class WVcrGameChooserWindow : public UIWindow {
        WGamedirChooser    dir_chooser;
        WVcrGameChooser    game_chooser;
        ObjectHolder       holder;
     public:
        WVcrGameChooserWindow(Ptr<Directory> dir, bool oneup);
        ~WVcrGameChooserWindow();
        bool handleEvent(const UIEvent& e, bool second_pass);
        void maybeStop(int);
        void switchTo(int id);
    };
}

/** Initialize.
    \param id      widget id
    \param gdc     associated game directory chooser */
WVcrGameChooser::WVcrGameChooser(int id, WGamedirChooser* gdc)
    : WGameRaceChooser(id), gdc(gdc)
{
    gdc->sig_moved.add(this, &WVcrGameChooser::handleMovement);
}

void
WVcrGameChooser::scan(Ptr<Directory> dir)
{
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        Ptr<Stream> s;

#ifdef FLAK_SUPPORT
        using namespace std;
        file_name[i-1] = format("flak%d.dat", i);
        s = dir->openFileNT(file_name[i-1], Stream::C_READ);
        TFlakHeader header;
        if (s && getStructure(*s, header) && memcmp(header.magic, FLAK_MAGIC, sizeof(header.magic)) == 0
            && header.num_battles > 0)
        {
            setInfo(i, true, format(_("%d fight%!1{s%}"), header.num_battles));
            continue;
        }
#endif

        file_name[i-1] = format("vcr%d.dat", i);
        s = dir->openFileNT(file_name[i-1], Stream::C_READ);
        TInt16 value;
        if (s && getStructure(*s, value) && value.value >= 0) {
            setInfo(i, true, format(_("%d fight%!1{s%}"), value.value));
        }
    }
}

/** Update after directory changed. */
void
WVcrGameChooser::handleMovement(int)
{
    setDirectory(gdc->getSubdirectory());
}

/** Initialize.
    \param dir     initial directory
    \param oneup   false: display directory as is;
                   true: display dir's parent, place cursor on dir */
WVcrGameChooserWindow::WVcrGameChooserWindow(Ptr<Directory> dir, bool oneup)
    : UIWindow(_("Choose Game"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      dir_chooser(dir, oneup, 10),
      game_chooser(11, &dir_chooser)
{
    setLayout(UIVBoxLayout::instance5);
    setState(st_Modal, false);

    UIGroup& v      = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& v2     = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& h      = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& h2     = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIButton& cd    = holder.add(new UIButton(0, _("Change directory"), 0, UIButton::aNothing));
    UIButton& watch = holder.add(new UIButton(0, _("View VCRs"), 0, UIButton::aNothing));
    // Borland sucks.
    void (WGamedirChooser::*tmp)(int) = &WGamedirChooser::enter;
    cd.sig_fired.add(&dir_chooser, tmp);

    watch.sig_fired.add(this, &WVcrGameChooserWindow::maybeStop);
    game_chooser.sig_dblclick.add(this, &WVcrGameChooserWindow::maybeStop);

    h.add(cd);
    h.add(holder.add(new UISpacer()));
    h.add(watch);
    h.add(holder.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE, UIButton::aStop)));

    v.add(holder.add(new UIStaticText(_("Race:"), UIStaticText::BlackTitle)));
    v.add(game_chooser);
    v.add(holder.add(new UISpacer()));

    v2.add(holder.add(new UIStaticText(_("Game directory:"), UIStaticText::BlackTitle)));
    v2.add(dir_chooser);

    h2.add(v2);
    h2.add(v);

    add(h2);
    add(h);

    pack();
}

WVcrGameChooserWindow::~WVcrGameChooserWindow()
{ }

/** Switch to widget. Switches to widget by Id, if possible. */
void
WVcrGameChooserWindow::switchTo(int id)
{
    UIWidget* w = getWidgetById(id);
    if (w && !w->hasState(st_Disabled))
        w->setFocus();
}

static void
picturePlay()
{
    Ptr<GfxPixmap> pix = requestPixmap("playvcr");
    ASSERT(pix);
    pix = pix->convertToScreenFormat();

    int coords[2][2] = { { 50, 0 }, { 0, 50 } };
    int index = 0;

    while (1) {
        /* display it */
        ui_screen->lock();
        drawSolidBar(*ui_screen, GfxRect(ui_screen->getSystemSurface()), standard_colors[COLOR_GRAY]);
        doStretchRotateBlit(pix->getSystemSurface(),
                            ui_screen->getSystemSurface(),
                            GfxRect(pix->getSystemSurface()),
                            GfxRect(ui_screen->getSystemSurface()),
                            320, 240,
                            coords[0][0], coords[0][1],
                            coords[1][0], coords[1][1]);
        ui_screen->unlock();
        ui_screen->update(GfxRect(ui_screen->getSystemSurface()));

        UIEvent e;
        do
            getEvent(e, false);
        while (e.type != ev_Key);
        switch (e.key.code) {
         case SDLK_TAB:
            index ^= 1;
            break;
         case SDLK_LEFT:
            coords[index][0] -= 10;
            break;
         case SDLK_RIGHT:
            coords[index][0] += 10;
            break;
         case SDLK_UP:
            coords[index][1] -= 10;
            break;
         case SDLK_DOWN:
            coords[index][1] += 10;
            break;
         case SDLK_ESCAPE:
            ui_screen->lock();
            ui_root->redraw();
            ui_screen->unlock();
            return;
        }
    }
}

/** Event handler. Handles Tab/Left/Right and Enter keys. */
bool
WVcrGameChooserWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    UIWidget* w = getFocusedWidget();
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
            if (w && w->getId() == 10)
                switchTo(11);
            else
                switchTo(10);
            return true;
         case SDLK_RIGHT:
            switchTo(11);
            return true;
         case SDLK_LEFT:
            switchTo(10);
            return true;
         case SDLK_RETURN:
            if (w && w->getId() == 10) {
                dir_chooser.enter();
            } else {
                maybeStop(0);
            }
            return true;
         case SDLK_F1:
            picturePlay();
            return true;
         case vk_Quit:
            stop(cm_Escape);
            putKey(vk_Quit);
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

/** Handler for "View VCRs" button. If possible, set parameters as
    entered in the dialog and stop() it. */
void
WVcrGameChooserWindow::maybeStop(int)
{
    if (Ptr<Directory> d = dir_chooser.getSubdirectory()) {
        if (game_chooser.isAnyRaceAccessible()) {
            PlayvcrOptionParser& parser = PlayvcrOptionParser::Instance;
            parser.setDirName(d->getName());
            parser.setFileName(Directory::makeFileName(d->getName(),
                                                       game_chooser.getFileName(game_chooser.getCurrent()+1)));
            stop(cm_Enter);
        }
    }
}

/** Directory chooser, main interface.
    \returns true iff user chose "OK" */
bool
chooseDirectories(Ptr<Directory> dir, bool oneup)
{
    WVcrGameChooserWindow w(dir, oneup);
    ui_root->centerWidget(w);
    ui_root->add(w);
    int rv = ui_root->run();
    ui_root->close(w);
//    optDirectory = true;

    return (rv == cm_Enter);
}

static bool
changeDirectory()
{
    return chooseDirectories(game_file_dir, true);
}

/** Handler for "File" button. Change directory, and stop VCR dialog if
    user chose "OK" */
static void
changeDirectoryAction(int)
{
    if (changeDirectory())
        ui_root->stop(1);
}

static const UIKeyTranslator::Init VCR_SEL_KEYS[] = {
    { SDLK_UP,      cm_VCR_Prev },
    { SDLK_DOWN,    cm_VCR_Next },
    { vk_WheelDown, cm_VCR_Next },
    { vk_WheelUp,   cm_VCR_Prev },
    { vk_Quit,      cm_Escape }
};

/** Play VCRs.
    \returns true if user chose a different file; main code should reload
    data and restart playing. */
static bool
playVcrs(GClassicVcrDatabase& db, int index)
{
    ObjectHolder holder;
    UIWindow w(_("Visual Combat Recorder"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIButton f(1, _("F - File"), 'f', UIButton::aNothing);
    f.sig_fired.add(changeDirectoryAction);

    UIGroup but(UIHBoxLayout::instance5);
    but.add(holder.add(new UIButton(cm_VCR_Prev, "\036", '-')));
    but.add(holder.add(new UIButton(cm_VCR_Next, "\037", '+')));
    but.add(holder.add(new UIButton(cm_Enter, _("ENTER - Play"), SDLK_RETURN)));
    but.add(holder.add(new UISpacer()));
    but.add(f);
    but.add(holder.add(new UIButton(cm_Escape, _("ESC - Exit"), SDLK_ESCAPE, UIButton::aStop)));

    WVcrSelector sel(db);
    if (index)
        sel.selectVcr(index-1);

    w.setLayout(UIVBoxLayout::instance5);
    w.setState(UIWindow::st_Modal, false);
    w.add(sel);
    w.add(but);
    w.add(holder.add(new UIKeyTranslator(VCR_SEL_KEYS, countof(VCR_SEL_KEYS))));
    w.pack();
    ui_root->centerWidget(w);
    ui_root->add(w);
    if (index)
        putCommand(cm_Enter, 0);
    bool rv = (ui_root->run() != cm_Escape);
    ui_root->close(w);

    return rv;
}

#ifdef FLAK_SUPPORT
namespace {
    class FlakVcrSelectorWithHandler : public FlakVcrSelector {
     public:
        FlakVcrSelectorWithHandler(GFlakVcrDatabase& db)
            : FlakVcrSelector(db) { }
        bool handleEvent(const UIEvent& e, bool second_pass);
    };
}

string_t getFlakLog();

bool
FlakVcrSelectorWithHandler::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Command) {
        switch (e.cmd.code) {
         case cm_VCR_Next:
            next(e.cmd.info);
            return true;
         case cm_VCR_Prev:
            previous(e.cmd.info);
            return true;
         case cm_Enter:
            playFlakBattle(getDatabase().getBattle(getCurrent()));
            return true;
         case cm_VCR_LeftInfo:
            showInfo();
            return true;
        }
    }
    if (e.type == ev_Key) {
        if (e.key.code == 'i') {
            string_t s = getFlakLog();
            if (s.size()) {
                messageBox(s, "FLAK Message Log");
            }
            return true;
        } else if (e.key.code == 'z') {
            // HACK
            const FlakVcrEntry& ve = getDatabase().getBattle(getCurrent());
            FlakBattle& b = const_cast<FlakBattle&>(ve.getBattle());
            uint32 saved_seed = b.getOrigSeed();
            b.setOrigSeed(rand());
            playFlakBattle(getDatabase().getBattle(getCurrent()));
            b.setOrigSeed(saved_seed);
            return true;
        }
    }
    return FlakVcrSelector::handleEvent(e, second_pass);
}

static bool
playFlak(GFlakVcrDatabase& db, int index)
{
    ObjectHolder h;

    UIWindow w(_("Visual Combat Recorder"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIButton f(1, _("F - File"), 'f', UIButton::aNothing);
    f.sig_fired.add(changeDirectoryAction);

    UIGroup but(UIHBoxLayout::instance5);
    but.add(h.add(new UIButton(cm_VCR_Prev, "\036", '-')));
    but.add(h.add(new UIButton(cm_VCR_Next, "\037", '+')));
    but.add(h.add(new UIButton(cm_Enter, _("ENTER - Play"), SDLK_RETURN)));
    but.add(h.add(new UISpacer()));
    but.add(f);
    but.add(h.add(new UIButton(cm_Escape, _("ESC - Exit"), SDLK_ESCAPE, UIButton::aStop)));

    UIGroup sel1(UIHBoxLayout::instance5), sel2(UIVBoxLayout::instance5);

    FlakVcrSelectorWithHandler sel(db);
    if (index)
        sel.setCurrent(index - 1);
    sel1.add(sel);
    sel1.add(sel2);
    sel2.add(h.add(new UIButton(cm_VCR_LeftInfo, "L", 'l')));
    sel2.add(h.add(new UISpacer()));

    w.setLayout(UIVBoxLayout::instance5);
    w.setState(UIWindow::st_Modal, false);
    w.add(sel1);
    w.add(but);
    w.add(h.add(new UIKeyTranslator(VCR_SEL_KEYS, countof(VCR_SEL_KEYS))));
    w.pack();

    ui_root->centerWidget(w);
    ui_root->add(w);
    if (index)
        putCommand(cm_Enter, 0);
    bool rv = (ui_root->run() != cm_Escape);
    ui_root->close(w);

    return rv;
}
#endif

static void
doSoundConfig(int)
{
    SfxConfig conf = getSoundConfig();
    if (configureSoundDialog(conf)) {
        configureSound(conf);
        saveSoundConfig();
    }
}

static void
doGeneralConfig(int)
{
    ObjectHolder holder;
    UICheckbox cb_instant(100, _("Instant battle result"), 'i', 2);
    UIGroup h(UIHBoxLayout::instance5);

    UIWindow win(_("General Configuration"),
                 UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);

    h.add(holder.add(new UISpacer()));
    h.add(holder.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN, UIButton::aStop)));
    h.add(holder.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE, UIButton::aStop)));

    win.add(cb_instant);
    win.add(h);
    win.pack();

    cb_instant.setValue(user_pref.InstantBattleResult);
    cb_instant.setFocus();

    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);

    if (rv == cm_Enter) {
        user_pref.InstantBattleResult = cb_instant.getValue();
        ui_root->getFirstWidget()->drawWidget(false);
        config_file.setValue("vcr.instantresult", user_pref.InstantBattleResult);
    }
}

#ifdef SIMULATOR_SUPPORT
static void
doSimulator()
{
    GSimulatorNullInterface null_if;
    doSimulator(null_if);
}
#endif

class PlayVcrVisitor : public GVcrDatabaseVisitor {
    int index;
    bool rv;
 public:
    PlayVcrVisitor(int index)
        : index(index) { }
    void visitClassicDatabase(GClassicVcrDatabase& db)
        {
            rv = playVcrs(db, index);
        }
    void visitFlakDatabase(GFlakVcrDatabase& db)
        {
#ifdef FLAK_SUPPORT
            rv = playFlak(db, index);
#endif
        }
    bool doContinue() const
        { return rv; }
};

int
main(int, char** argv)
{
    Platform::initOutput();
    initNLS();
    PlayvcrOptionParser& parser = PlayvcrOptionParser::Instance;
    Platform::initOptions(parser);
    initGraphicsOptions(parser);
    parser.parse(argv);
#ifdef FLAK_SUPPORT
    SDL_Surface* sfc = initGraphics("playvcr.bmp", _("PCC II VCR/FLAK Player"));
#else
    SDL_Surface* sfc = initGraphics("playvcr.bmp", _("PCC II VCR Player"));
#endif
    if (!sfc) {
        Platform::showOutput();
        return 1;
    }

    initDebug();
    try {
        ObjectHolder global_wids;

        loadConfig();
        initUIGlobals(new GfxSurface(sfc, false),
                      Directory::openDir(Platform::getResourceDirectory()));
        loadSoundConfig();          // this will also initialize sound
        config_file.getValue("vcr.instantresult", user_pref.InstantBattleResult);
        {
            GfxRect r = ui_root->getExtent();
            r.h -= 200;
            r.y += 200;

            UIWidget& w = global_wids.add(new PlayvcrConsole());
            w.setExtent(r);
            ui_root->add(w);

            UIAbstractButton& b = global_wids.add(new PlayvcrSimpleButton(0, _("F2 - Sound Config"), SDLK_F2));
            int y = 0;
            UIWidget::LayoutInfo li;
            b.getLayoutInfo(li);
            b.sig_fired.add(doSoundConfig);
            b.setExtent(GfxRect(r.w - li.pref_size.x, y, li.pref_size.x, li.pref_size.y));
            ui_root->add(b);
            y += li.pref_size.y + 1;

            UIAbstractButton& b2 = global_wids.add(new PlayvcrSimpleButton(0, _("F3 - General Config"), SDLK_F3));
            b2.getLayoutInfo(li);
            b2.sig_fired.add(doGeneralConfig);
            b2.setExtent(GfxRect(r.w - li.pref_size.x, y, li.pref_size.x, li.pref_size.y));
            ui_root->add(b2);
            y += li.pref_size.y + 1;

#ifdef SIMULATOR_SUPPORT
            UIAbstractButton& b3 = global_wids.add(new PlayvcrSimpleButton(0, _("F4 - Simulator"), SDLK_F4));
            b3.getLayoutInfo(li);
            b3.sig_fired.add(doSimulator);
            b3.setExtent(GfxRect(r.w - li.pref_size.x, y, li.pref_size.x, li.pref_size.y));
            ui_root->add(b3);
#endif

#ifdef FLAK_SUPPORT
            Ptr<GfxPixmap> pix = requestPixmap("vcr.flak");
#else
            Ptr<GfxPixmap> pix = requestPixmap("vcr.banner");
#endif
            if (pix)
                ui_root->add(global_wids.add(new PlayvcrImageWidget(pix)));
        }

        if (parser.getNumNonoptions() == 0 || optDirectory) {
            /* we need directories for initial race names; we need
               config and specification in case user calls simulator */
            initDirectories(Platform::getDataDirectory(), Platform::getDataDirectory());
            initConfig();
            initSpecification();
            initRaceNames();
            if (!chooseDirectories((parser.getNumNonoptions()
                                    ? Directory::openDir(parser.getFileName())
                                    : Directory::openDir(Directory::getWorkingDirectory())),
                                   false))
                goto skip;
        }

        bool first;
        first = true;
        while (1) {
            initDirectories(parser.getDirName(), Platform::getDataDirectory());
            initConfig();
            initSpecification();
            initRaceNames();

            Ptr<Stream> str = Stream::openFileT(parser.getFileName(), Stream::C_READ);
            Ptr<GVcrDatabase> db = 0;
#ifdef FLAK_SUPPORT
            TFlakHeader hdr;
            if (getStructure(*str, hdr) && std::memcmp(hdr.magic, FLAK_MAGIC, sizeof(hdr.magic)) == 0) {
                str->seek(0);
                db = new GFlakVcrDatabase(*str);
            }
#endif
            if (!db) {
                str->seek(0);
                db = new GClassicVcrDatabase(*str);
            }
            str = 0;
            
            console.write(LOG_INFO, format(_("Loaded %d record%!1{s%} from `%s'"),
                                           db->getCount(), parser.getFileName()));
            if (!db->getCount()) {
                messageBox(format(_("The file `%s' does not contain any battles."),
                                  parser.getFileName()),
                           _("VCR Player"));
                if (!changeDirectory())
                    break;
            } else {
                int index = first ? parser.getBattleNumber() : 0;
                if (index && index > db->getCount()) {
                    messageBox(format(_("There are only %d battles in this VCR file."), db->getCount()),
                               _("VCR Player"));
                    index = 0;
                }

                PlayVcrVisitor vis(index);
                db->visit(vis);
                if (!vis.doContinue())
                    break;
            }
        }
    skip:
        ui_root = 0;
        ui_screen = 0;

        saveConfig();
    }
    catch (std::exception& e) {
        console << _("Error during initialisation: ") << e.what() << "\n";
        Platform::showOutput();
    }
    return 0;
}
