/**
  *  \file client/winfo.cc
  *  \brief Ship Component Info Windows
  *
  *  This file defines a few widgets that display information about a
  *  starship component.
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include "client/winfo.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/player.h"
#include "game/spec.h"
#include "gfx/font.h"
#include "ui/skin.h"
#include "ui/table.h"
#include "ui/uiglobal.h"
#include "util/misc.h"
#include "util/nls.h"
#include "game/host.h"

/// Number of lines for hull functions.
enum { HULLINFO_LINES = 10 };

/******************************* WHullInfo *******************************/

/** \class WHullInfo
    \brief Information about a Hull */

/** Constructor. */
WHullInfo::WHullInfo()
    : UIWidget(), hull_id(0)
{ }

WHullInfo::~WHullInfo()
{ }

/** Set the hull to display.
    \param n Hull number */
void
WHullInfo::setHullId(int n)
{
    if (hull_id != n) {
        hull_id = n;
        drawWidget();
    }
}

void
WHullInfo::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UITableWriter tw(can, getSkin(), getExtent(), FONT_NORMAL, UITableWriter::fl_Fill);

    const char* const torp_str    = _("\7 Torpedoes:");
    const char* const fighter_str = _("\7 Fighters:");
    const char* const beam_str    = _("\7 Beams:");

    int min_col_width = std::max(fonts[FONT_NORMAL]->getTextWidth(torp_str),
                                 std::max(fonts[FONT_NORMAL]->getTextWidth(beam_str),
                                          fonts[FONT_NORMAL]->getTextWidth(fighter_str)));

    tw.newLColumn(min_col_width);
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Mass:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Cargo:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Fuel:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Engines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Crew:"));

    if (hull_id) {
        GHull& h = getHull(hull_id);
        int num_blank = 2; /* engines, crew */

        /* This is a hack - sort of. UITableWriter does not support
           multi-column entries yet. Because it does not expect us to
           write the hull functions as multi-column entries, it will
           not correctly overwrite them and leave dirt around. */
        GfxPoint pt = tw.getCurrentPosition();
        getSkin().drawBackground(can, GfxRect(pt.x, pt.y, extent.x + extent.w - pt.x, extent.y + extent.h - pt.y));

        if (h.getMaxBeams())
            tw.writeVariableCell(UIColor::tc_Static, beam_str), ++num_blank;
        if (h.getMaxLaunchers())
            tw.writeVariableCell(UIColor::tc_Static, torp_str), ++num_blank;
        if (h.getNumBays())
            tw.writeVariableCell(UIColor::tc_Static, fighter_str), ++num_blank;

        tw.newRColumn(4 * font_ems[FONT_NORMAL]);
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getMass()));
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getCargo()));
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getFuel()));
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getNumEngines()));
        tw.writeVariableCell(UIColor::tc_Green, numToString(h.getCrew()));
        if (h.getMaxBeams())
            tw.writeVariableCell(UIColor::tc_Green, numToString(h.getMaxBeams()));
        if (h.getMaxLaunchers())
            tw.writeVariableCell(UIColor::tc_Green, numToString(h.getMaxLaunchers()));
        if (h.getNumBays())
            tw.writeVariableCell(UIColor::tc_Green, numToString(h.getNumBays()));

        tw.newLColumn();
        tw.writeFixedCell(UIColor::tc_Green, "kt");
        tw.writeFixedCell(UIColor::tc_Green, "kt");
        tw.writeFixedCell(UIColor::tc_Green, "kt");
        do
            tw.writeFixedCell(UIColor::tc_Green, string_t());
        while (--num_blank);
        tw.finish();

        /* build up list of special devices */
        // FIXME: this code is horrible, and it better should be elsewhere
        std::vector<int> devices;
        std::vector<char> hidden;
        for (int func = 0; func < hf_MAX; ++func) {
            if (h.canDoSpecial(func, getPlayerId(), true) || h.canDoSpecial(func, getPlayerId(), false)) {
                devices.push_back(func);
                hidden.push_back(false);
            }
        }
        for (std::size_t index = 0; index < devices.size(); ++index) {
            GHullFunction* hf = GHullFunction::getHullFunctionById(devices[index]);
            if (hf && hf->getImpliedDevice() >= 0) {
                for (std::size_t other = 0; other < devices.size(); ++other)
                    if (devices[other] == hf->getImpliedDevice())
                        hidden[other] = true;
            }
        }

        int line = tw.getCurrentLine();
        getSkin().getParameters().just_x =
            getSkin().getParameters().just_y = 0;
        getSkin().setSolid(false); // not needed
        getSkin().setColor(UIColor::tc_Static);

        for (std::size_t i = 0; i < devices.size() && line < HULLINFO_LINES; ++i) {
            if (hidden[i])
                continue;

            GHullFunction* ptr = GHullFunction::getHullFunctionById(devices[i]);
            if (ptr) {
                getSkin().outTextF(can, FONT_NORMAL, getExtent().x,
                                   getExtent().y + line * font_heights[FONT_NORMAL],
                                   getExtent().w,
                                   "\7 " + ptr->getDescription());
                ++line;
            }
        }
    }

    ui.update(getExtent());
}

void
WHullInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = GfxPoint(20 * font_ems[FONT_NORMAL],
                                          HULLINFO_LINES * font_heights[FONT_NORMAL]);
}

/******************************* WBeamInfo *******************************/

/*! \class WBeamInfo
    \brief Information about a beam */

/** Constructor. */
WBeamInfo::WBeamInfo()
    : UIWidget(), beam_id(1)
{ }

WBeamInfo::~WBeamInfo()
{ }

/** Get beam classification. */
static const char*
getBeamType(const GBeam& b)
{
    /* translators: PHost 4 distinguishes between "death rays" and
       normal beams. We invent three new kinds for a change. Invent
       new names if you want. */
    if (GHost::hasDeathRays() && b.getDamage() == 0)
        return _("death ray");
    else if (b.getDamage() >= 2 * b.getKill())
        return _("atomizer beam");
    else if (b.getKill() >= 2 * b.getDamage())
        return _("X-ray");
    else
        return _("energy beam");
}

void
WBeamInfo::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UITableWriter tw(can, getSkin(), getExtent(), FONT_NORMAL, UITableWriter::fl_Fill);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Type:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Kill:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Destroy:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Recharge:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Hit:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Sweep:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Beam Mass:"));

    GBeam& b = getBeam(beam_id);

    tw.newRColumn(3 * font_ems[FONT_NORMAL]);
    GfxPoint mark = tw.getCurrentPosition();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, numToString(b.getKill()));
    tw.writeVariableCell(UIColor::tc_Green, numToString(b.getDamage()));

    if (host.isTHost()) {
        tw.writeVariableCell(UIColor::tc_Green, "100");
        tw.writeVariableCell(UIColor::tc_Green, "100");
    } else {
        /* Recharge time:
             THost: recharges 0.5 per tick, fires at 50  --> 100
             PHost: recharges by [0, (Kill+Damage) * BeamRechargeBonus + BeamRechargeRate)
               per tick, fires at BeamHitShipCharge
                -> 2*BeamHitShipCharge / ((Kill+Damage) * BRB + BRR)  */
        int bang = b.getKill() + b.getDamage();
        int recharge_rate = bang * CONFIG_ARRAY(BeamRechargeBonus, ::getPlayerId()) / 100 + CONFIG_ARRAY(BeamRechargeRate, ::getPlayerId());
        if (!recharge_rate)
            tw.writeVariableCell(UIColor::tc_Green, "-");
        else
            tw.writeVariableCell(UIColor::tc_Green, numToString(2 * CONFIG_ARRAY(BeamHitShipCharge, ::getPlayerId()) / recharge_rate));

        int hit = bang * CONFIG_ARRAY(BeamHitBonus, ::getPlayerId()) / 100 + CONFIG_ARRAY(BeamHitOdds, ::getPlayerId());
        if (hit < 0)
            hit = 0;
        if (hit > 100)
            hit = 100;
        tw.writeVariableCell(UIColor::tc_Green, numToString(hit));
    }

    tw.writeVariableCell(UIColor::tc_Green, numToString(beam_id * beam_id * CONFIG_ARRAY(MineSweepRate, ::getPlayerId())));
    tw.writeVariableCell(UIColor::tc_Green, numToString(b.getMass()));

    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, "s");
    tw.writeVariableCell(UIColor::tc_Green, "%");
    if (CONFIG_ARRAY(WebMineSweepRate, ::getPlayerId() == CONFIG_ARRAY(MineSweepRate, ::getPlayerId())))
        tw.writeVariableCell(UIColor::tc_Green, _("mines"));
    else
        tw.writeVariableCell(UIColor::tc_Green, format(_("mines, %d webs"),
                                                       beam_id * beam_id * CONFIG_ARRAY(WebMineSweepRate, ::getPlayerId())));
    tw.writeVariableCell(UIColor::tc_Green, "kt");
    tw.finish();

    GfxParameters& parm = getSkin();
    parm.just_x = parm.just_y = 0;
    getSkin().setSolid();
    getSkin().setColor(UIColor::tc_Green);
    getSkin().outTextF(can, FONT_NORMAL,
                       mark.x, mark.y, getExtent().x + getExtent().w - mark.x,
                       getBeamType(b));
    ui.update(getExtent());
}

void
WBeamInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = GfxPoint(20 * font_ems[FONT_NORMAL],
                                          5 * font_heights[FONT_NORMAL]);
}

void
WBeamInfo::setBeamId(int n)
{
    if (n != beam_id) {
        beam_id = n;
        drawWidget();
    }
}

/******************************* WTorpInfo *******************************/

/*! \class WTorpInfo
    \brief Information about a torpedo */
WTorpInfo::WTorpInfo()
    : UIWidget(), torp_id(1)
{ }

WTorpInfo::~WTorpInfo()
{ }

/** Get torpedo classification. */
static const char*
getTorpType(const GTorpLauncher& b)
{
    /* translators: PHost 4 distinguishes between "death rays" and
       normal torpedoes. We invent three new kinds for a change. 
       Invent new names if you want. */
    if (GHost::hasDeathRays() && b.getDamage() == 0)
        return _("death ray torp");
    else if (b.getDamage() >= 2 * b.getKill())
        return _("atomizer torp");
    else if (b.getKill() >= 2 * b.getDamage())
        return _("X-ray torp");
    else
        return _("explosive torp");
}

void
WTorpInfo::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UITableWriter tw(can, getSkin(), getExtent(), FONT_NORMAL, UITableWriter::fl_Fill);

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Type:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Kill:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Destroy:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Recharge:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Hit:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Torp Cost:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 1000 Mines:"));
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Launcher Mass:"));

    GTorpLauncher& l = getTorp(torp_id);
    int scale = (host.isTHost() || !config.AllowAlternativeCombat) ? 2 : 1;

    tw.newRColumn(3 * font_ems[FONT_NORMAL]);
    GfxPoint mark = tw.getCurrentPosition();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, numToString(scale * l.getKill()));
    tw.writeVariableCell(UIColor::tc_Green, numToString(scale * l.getDamage()));

    if (host.isTHost()) {
        /* These values directly follow from the Tim-VCR code,
           although they are never explicitly used as such. */
        static const int8 recharge_times[] = {
           41, 41, 36, 34, 33, 33, 32, 32, 32, 31
        };
        tw.writeVariableCell(UIColor::tc_Green, numToString(recharge_times[torp_id-1]));
        tw.writeVariableCell(UIColor::tc_Green, "66");
    } else {
        int bang = l.getKill() + l.getDamage();
        int rate = bang * CONFIG_ARRAY(TubeRechargeBonus, ::getPlayerId()) / 100 + CONFIG_ARRAY(TubeRechargeRate, ::getPlayerId());
        if (rate)
            tw.writeVariableCell(UIColor::tc_Green, numToString(2000 / rate));
        else
            tw.writeVariableCell(UIColor::tc_Green, "-");
        rate = bang * CONFIG_ARRAY(TorpHitBonus, ::getPlayerId()) / 100 + CONFIG_ARRAY(TorpHitOdds, ::getPlayerId());
        if (rate < 0)
            rate = 0;
        if (rate > 100)
            rate = 100;
        tw.writeVariableCell(UIColor::tc_Green, numToString(rate));
    }

    tw.writeVariableCell(UIColor::tc_Green, numToString(l.getTorpCost()));

    /* how many mines do we get from 100 torps? */
    int mine_lay_divi = (host.isPHost()
                         ? CONFIG_ARRAY(UnitsPerTorpRate, ::getPlayerId())
                         : raceId(::getPlayerId()) == 9
                         ? 400 : 100)
        * torp_id * torp_id;

    /* simple equation: if 100 torps give mine_lay_divi mines, how
       many torps do we need for 1000 mines? */
    if (mine_lay_divi <= 0)
        tw.writeVariableCell(UIColor::tc_Green, "-");
    else
        tw.writeVariableCell(UIColor::tc_Green, numToString(100000L * l.getTorpCost() / mine_lay_divi));
    tw.writeVariableCell(UIColor::tc_Green, numToString(l.getMass()));

    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    tw.writeVariableCell(UIColor::tc_Green, "s");
    tw.writeVariableCell(UIColor::tc_Green, "%");
    tw.writeVariableCell(UIColor::tc_Green, _("mc, 1 T/D/M"));
    tw.writeVariableCell(UIColor::tc_Green, format("mc, %d T/D/M", 100000 / mine_lay_divi));
    tw.writeVariableCell(UIColor::tc_Green, "kt");
    tw.finish();

    GfxParameters& parm = getSkin();
    parm.just_x = parm.just_y = 0;
    getSkin().setSolid();
    getSkin().setColor(UIColor::tc_Green);
    getSkin().outTextF(can, FONT_NORMAL,
                       mark.x, mark.y, getExtent().x + getExtent().w - mark.x,
                       getTorpType(l));
    ui.update(getExtent());
}

void
WTorpInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = GfxPoint(20 * font_ems[FONT_NORMAL],
                                          5 * font_heights[FONT_NORMAL]);
}

void
WTorpInfo::setTorpId(int n)
{
    if (n != torp_id) {
        torp_id = n;
        drawWidget();
    }
}

/****************************** WEngineInfo ******************************/

WEngineInfo::WEngineInfo()
    : UIWidget(), engine_id(1)
{ }

WEngineInfo::~WEngineInfo()
{ }

void
WEngineInfo::draw(GfxCanvas& can, UpdateInfo& ui)
{
    UITableWriter tw(can, getSkin(), getExtent(), FONT_NORMAL, UITableWriter::fl_Fill);
    GEngine& e = getEngine(engine_id);

    int max_warp = 0;
    do
        ++max_warp;
    while (max_warp <= 9 && e.getFuelFactor(max_warp) < 120L * max_warp * max_warp);
    --max_warp;

    tw.newLColumn();
    tw.writeFixedCell(UIColor::tc_Static, _("\7 Max efficient Warp:"));
    if (config.AllowEngineShieldBonus)
        tw.writeFixedCell(UIColor::tc_Static, _("\7 Shield Bonus:"));

    tw.newRColumn(3 * font_ems[FONT_NORMAL]);
    tw.writeVariableCell(UIColor::tc_Green, numToString(max_warp));
    if (config.AllowEngineShieldBonus)
        tw.writeVariableCell(UIColor::tc_Green,
                             numToString(e.getCost().mc * config.EngineShieldBonusRate / 100));

    tw.newLColumn();
    tw.writeVariableCell(UIColor::tc_Green, string_t());
    if (config.AllowEngineShieldBonus)
        tw.writeFixedCell(UIColor::tc_Green, "kt");

    tw.finish();
    ui.update(getExtent());
}

void
WEngineInfo::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.min_size = li.pref_size = GfxPoint(20 * font_ems[FONT_NORMAL],
                                          5 * font_heights[FONT_NORMAL]);
}

void
WEngineInfo::setEngineId(int n)
{
    if (n != engine_id) {
        engine_id = n;
        drawWidget();
    }
}
