/**
  *  \file game/act-supplies.cc
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/action.h"
#include "cpluslib/assert.h"
#include "util/nls.h"

/*********************** GPlanetSellSuppliesAction ***********************/

/*! \class GPlanetSellSuppliesAction
    \brief Supply Sale

    This transaction allows you to sell supplies.

    \todo This is a very simple straightforward implementation.
    It does not support undo nor meaningful concurrency. */

/** Constructor.
    \param the_planet planet
    \pre we are playing the planet */
GPlanetSellSuppliesAction::GPlanetSellSuppliesAction(GPlanet& the_planet)
    : the_planet(the_planet), transfer(the_planet)
{
    // can we do it?
    if (!the_planet.isPlayed())
        throw GError(GError::eNotPlaying, _("You do not own this planet"));

    // reserve everything for us
    reserved_sup = transfer.removeCargo(el_Supplies,
                                        transfer.getCurrent(el_Supplies),
                                        true);
    amount = 0;
}

/** Destructor. */
GPlanetSellSuppliesAction::~GPlanetSellSuppliesAction()
{
    if (!isFinished())
        doCancel();
    ASSERT(reserved_sup == 0);
}

/** Get minimum amount.
    FIXME: always returns 0; later this would be negative to allow
    undo-supply-sale. */
int32
GPlanetSellSuppliesAction::getMinimum() const
{
    return 0;
}

/** Get maximum amount to sell. */
int32
GPlanetSellSuppliesAction::getMaximum() const
{
    return reserved_sup;
}

/** Set amount to sell. Out-of-range values are automatically
    adjusted. */
void
GPlanetSellSuppliesAction::setAmount(int32 amount)
{
    if (amount < getMinimum())
        amount = getMinimum();
    if (amount > getMaximum())
        amount = getMaximum();
    if (amount != this->amount) {
        this->amount = amount;
        sig_changed.raise();
    }
}

void
GPlanetSellSuppliesAction::commit()
{
    // commit:
    // - put back supplies
    // - add money
    // FIXME: when we have bought supplies, amount will be negative. we
    // have to do something here.
    // FIXME: addCargo could possibly return failure. what do we do?
    transfer.addCargo(el_Supplies, reserved_sup - amount, true);
    transfer.addCargo(el_Money,    amount, true);
    transfer.doCommit();

    reserved_sup = amount = 0;
}

void
GPlanetSellSuppliesAction::cancel()
{
    // cancel: give back all supplies we have reserved, and don't
    // change money. Actually, ignore this blurb and just cancel the
    // container. (Since the container is a subtransaction, not a peer,
    // we are allowed to cancel it.)
    transfer.doCancel();
    amount = reserved_sup = 0;
}
