/**
  *  \file game/act-tech.cc
  *  \brief Tech Upgrade rule class.
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/action.h"
#include "cpluslib/assert.h"
#include "game/player.h"

/*********************** GStarbaseTechUpgradeAction **********************/

/** \class GStarbaseTechUpgradeAction
    \brief Tech Upgrade rule class.

    This class encapsulates the rules for tech upgrades, and billing
    there-of. */

/** Constructor. Initializes the object and allocates resources.
    \param the_planet the planet we're talking about
    \param container  the cargo container through which we are billing
                      the transaction. */
GStarbaseTechUpgradeAction::GStarbaseTechUpgradeAction(GPlanet& the_planet,
                                                       GCargoContainer& container)
    : GStarbaseBuildAction(the_planet, container),
      the_planet(the_planet)
{
    if (the_planet.hasBase()) {
        for (int i = 0; i < 4; ++i)
            new_tech[i] = min_tech[i] = the_planet.getBaseTech(TTechLevel(i));
    } else {
        for (int i = 0; i < 4; ++i)
            new_tech[i] = min_tech[i] = 1;
    }
    ASSERT(getCost() == 0);
}

GStarbaseTechUpgradeAction::~GStarbaseTechUpgradeAction()
{ }

/** Check whether the user may raise his tech level to the specified
    value. Only checks rules, not resources. */
bool
GStarbaseTechUpgradeAction::isValid(int index, int new_val) const
{
    ASSERT(index >= 0 && index < 4);
    if (new_val < getMinTech(index) || new_val > getMaxTech(index))
        return false;
    return true;
}

/** Get maximum allowed tech level for this index. */
int
GStarbaseTechUpgradeAction::getMaxTech(int /*index*/) const
{
    if (allowRegistered(the_planet.getOwner()))
        return 10;
    else
        return 6;
}

/** Set tech level to specified value. This will not actually update
    the base, it will only register the request and update the cost
    accordingly. The actual update happens in commit().
    \return true iff request was honored (see isValid(int,int)). */
bool
GStarbaseTechUpgradeAction::setTech(int index, int new_val)
{
    if (isValid(index, new_val)) {
        new_tech[index] = new_val;
        regenerate();
        return true;
    } else {
        return false;
    }
}

void
GStarbaseTechUpgradeAction::doChanges(Actor& a)
{
    for (int i = 0; i < 4; ++i)
        a.setTech(TTechLevel(i), new_tech[i]);
}

