/**
  *  \file game/ship-form.cc
  *  \brief Ship Formulas and Prediction
  *
  *  This module contains functions to compute ship properties and
  *  predict their behaviour.
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <cmath>
#include "game/ship.h"
#include "game/mission.h"
#include "game/player.h"
#include "game/config.h"
#include "game/coord.h"
#include "game/planet.h"
#include "game/formula.h"
#include "game/host.h"
#include "game/univ.h"

const int GShipTurnPredictor::MOVEMENT_TIME_LIMIT;

/** Constructor. Create a predictor for a single ship.
    \param univ   universe containing the ship
    \param id     ship Id */
GShipTurnPredictor::GShipTurnPredictor(GUniverse& univ, int id)
    : id(id), towee_override(0),
      univ(univ), fuel_used(0), turns_computed(0)
{
    univ.getShip(id).getRawData(this->ship);
    ship.owner = univ.getShip(id).getRealOwner();
}

/** Constructor. Create a predictor for a ship towing another ship.
    \param univ   universe containing the ship
    \param id     ship Id
    \param towee_override predictor for towed ship */
GShipTurnPredictor::GShipTurnPredictor(GUniverse& univ, int id, GShipTurnPredictor& towee_override)
    : id(id), towee_override(&towee_override),
      univ(univ), fuel_used(0), turns_computed(0)
{
    univ.getShip(id).getRawData(this->ship);
    ship.owner = univ.getShip(id).getRealOwner();
}

static void
doRefinery(TShip& s, int16& ore, int16& sup)
{
    int n = getHull(s.hull).getFuel() - s.ore_cargo[el_Neutronium];
    if (n > ore)
        n = ore;
    if (n > sup)
        n = sup;
    s.ore_cargo[el_Neutronium] += n;
    sup -= n;
    ore -= n;
}

static int
getCloakFuel(const TShip& s, int turns)
{
    int fuel = getHull(s.hull).getMass() * config.CloakFuelBurn / 100;
    if (fuel < config.CloakFuelBurn)
        fuel = config.CloakFuelBurn;
    if (turns)
        fuel *= turns;
    return fuel;
}

// FIXME: code duplication. We already have this in ship.cc (computeMass).
static int
getShipMass(const TShip* ship)
{
    int mass = getHull(ship->hull).getMass()
        + ship->ammo
        + ship->ore_cargo[0] + ship->ore_cargo[1]
        + ship->ore_cargo[2] + ship->ore_cargo[3]
        + ship->colonists + ship->supplies;
    if (ship->torp_launcher_count)
        mass += ship->torp_launcher_count * getTorp(ship->torp_type).getMass();
    if (ship->beam_count)
        mass += ship->beam_count * getBeam(ship->beam_type).getMass();
    return mass;
}

static int
getEngineLoad(GUniverse& univ, TShip& ship, TShip* towee_override, bool tow_corr)
{
    int mass = getShipMass(&ship);
    if (ship.mission == GMission::msn_Tow && ship.mission_tow > 0 && ship.mission_tow <= NUM_SHIPS) {
        int towee_mass = 0;
        if (towee_override && towee_override->ship_id == ship.mission_tow)
            towee_mass = getShipMass(towee_override);
        else if (univ.getShip(ship.mission_tow).isExistant())
            // FIXME: range check
            towee_mass = univ.getShip(ship.mission_tow).getMass();
        if (tow_corr)
            towee_mass = 10*(towee_mass/10);
        mass += towee_mass;
    }
    return mass;
}

static int
computeFuelUsage(GUniverse& univ, TShip& ship, TShip* towee_override,
                 bool grav_acc, double dist)
{
    if (ship.warp == 0)
        return 0;

    int way = ship.warp*ship.warp;
    if (grav_acc)
        way *= 2;

    int load = getEngineLoad(univ, ship, towee_override, host.isTHost());
    long ff = getEngine(ship.engine).getFuelFactor(ship.warp);

    if (host.isTHost()) {
        // THost formula
        return (ff * (load / 10) * long(dist)) / (10000L * way);
    } else if (!config.UseAccurateFuelModel) {
        // PHost, standard formula
        return rdivaddl(load, 10, 0) * ff * long(dist) / (10000L * way);
    } else {
        // PHost, "accurate" formula
        return roundToInt(load * (1 - std::exp(-(ff * dist) / (way*100000L))));
    }
}

static void
normalizePosition(TShip& ship)
{
    GPoint new_pos = GPoint(ship.x, ship.y).getNormalized();
    ship.x = new_pos.x;
    ship.y = new_pos.y;
}

/** Compute one turn. This will compute all changes for the ship. It
    will also update the towed ship predictor, if any. */
void
GShipTurnPredictor::computeTurn()
{
    // where are we?
    int pid                = univ.getPlanetAt(GPoint(ship.x, ship.y));
    const GShip& real_ship = univ.getShip(id);

//     {+++ Mine Laying +++}
//     IF (s.Neutronium<>0) AND (ComputeMineDrop(@s, md)) THEN BEGIN
//       Dec(s.Torps, md.torps);
//     END;

    // Special Missions I (Super Refit, Self Repair, Hiss, Rob) would go here

    /* Alchemy */
    if (std::memcmp(ship.fcode, "NAL", 3) != 0) {
        if (real_ship.canDoSpecial(hf_MerlinAlchemy)) {
            // Merlin converts 3 Sup -> 1 Min
            int t = 0, d = 0, m = 0;
            int mins = ship.supplies / 3;
            if (allowRegistered(ship.owner)) {
                if (std::memcmp(ship.fcode, "alt", 3) == 0)
                    t = mins;
                else if (std::memcmp(ship.fcode, "ald", 3) == 0)
                    d = mins;
                else if (std::memcmp(ship.fcode, "alm", 3) == 0)
                    m = mins;
                else if (host.isPHost()) {
                    if (std::memcmp(ship.fcode, "nat", 3) == 0)
                        d = m = mins/2;
                    else if (std::memcmp(ship.fcode, "nad", 3) == 0)
                        t = m = mins/2;
                    else if (std::memcmp(ship.fcode, "nam", 3) == 0)
                        t = d = mins/2;
                    else
                        t = d = m = mins/3;
                } else
                    t = d = m = mins/3;
            } else
                t = d = m = mins/3;
            ship.supplies -= 3*(t+d+m);
            ship.ore_cargo[el_Tritanium]  += t;
            ship.ore_cargo[el_Duranium]   += d;
            ship.ore_cargo[el_Molybdenum] += m;
        } else if (real_ship.canDoSpecial(hf_NeutronicRefinery)) {
            // Neutronic refinery converts 1 Sup + 1 Min -> 1 Neu
            doRefinery(ship, ship.ore_cargo[el_Tritanium],  ship.supplies);
            doRefinery(ship, ship.ore_cargo[el_Duranium],   ship.supplies);
            doRefinery(ship, ship.ore_cargo[el_Molybdenum], ship.supplies);
        } else if (real_ship.canDoSpecial(hf_AriesRefinery)) {
            // Aries converts 1 Min -> 1 Neu
            int16 max = 0x7FFF;  // we assume that no ship has more cargo
            doRefinery(ship, ship.ore_cargo[el_Tritanium],  max);
            doRefinery(ship, ship.ore_cargo[el_Duranium],   max);
            doRefinery(ship, ship.ore_cargo[el_Molybdenum], max);
        }
    }

    /* Starbase Missions I (Fix, [Recycle, Load Torps]) */
    if (pid > 0) {
        const GPlanet& p = univ.getPlanet(pid);
        if (p.isExistant()
            && p.hasBase()
            && p.getBaseShipyardOrder() == sya_Fix
            && p.getBaseShipyardId() == id)
        {
            // we're at a base which is fixing us
            ship.damage = 0;
            ship.crew   = getHull(ship.hull).getCrew();
        }
    }

    /* Supply Repair */
    if (ship.damage) {
        int max = ship.supplies / 5;
        if (max > ship.damage)
            max = ship.damage;
        ship.damage -= max;
        ship.supplies -= 5*max;
    }

    /* Cloak Fuel Burn */
    if ((real_ship.canDoSpecial(hf_Cloak) || real_ship.canDoSpecial(hf_AdvancedCloak))
        && isMissionCloaking(ship.mission, ship.owner))
    {
        int fuel = real_ship.canDoSpecial(hf_AdvancedCloak) ? 0 : getCloakFuel(ship, 0);
        if (ship.ore_cargo[el_Neutronium] <= fuel || ship.damage >= config.DamageLevelForCloakFail)
            ship.mission = 0;
        else
            ship.ore_cargo[el_Neutronium] -= fuel;
    }

    /* out of fuel cannot do any mission */
    if (ship.ore_cargo[el_Neutronium] == 0)
        ship.mission = 0;

    /* damage speed limit */
    if (ship.damage) {
        int limit = (raceId(ship.owner) == 2
                     ? (host.isPHost()
                        ? 15 - ship.damage/10
                        : 14 - ship.damage/10)
                     : 10 - ship.damage/10);
        if (limit < 0)
            limit = 0;
        if (ship.warp > limit)
            ship.warp = limit;
    }

    /* Actual movement here */
    long dist2 = long(ship.waypoint_dx*ship.waypoint_dx) + long(ship.waypoint_dy*ship.waypoint_dy);
    if (real_ship.canDoSpecial(hf_Hyperdrive)
        && std::memcmp(ship.fcode, "HYP", 3) == 0
        && ship.warp
        && dist2 >= 20*20)
    {
        /* It's hyperjumping */
        ship.ore_cargo[el_Neutronium] -= 50;
        fuel_used                     += 50;

        /* If it's jumping, it can't tow. Advance time in towee's
           world anyway. */
        if (ship.mission == GMission::msn_Tow)
            ship.mission = 0;
        if (towee_override)
            towee_override->computeTurn();

        /* Now move that bugger. */
        long foo = host.isPHost() ? 0 : 1;
        if (dist2-foo >= 340L*340 && dist2+foo <= 360L*360) {
            // exact jump. The obscure comparison is a creative
            // rendering of the fact that 360.00 and 340.00 are *NOT*
            // exact jumps in THost, but are in PHost.
            ship.x += ship.waypoint_dx;
            ship.y += ship.waypoint_dy;
        } else {
            // non-exact jump
            double dist = 350 / std::sqrt(double(dist2));
            int mx = int(dist * std::abs(ship.waypoint_dx) + 0.4999999);
            int my = int(dist * std::abs(ship.waypoint_dy) + 0.4999999);
            if (ship.waypoint_dx < 0)
                mx = -mx;
            if (ship.waypoint_dy < 0)
                my = -my;
            ship.x += mx;
            ship.y += my;
        }
        ship.waypoint_dx = ship.waypoint_dy = 0;
        ship.warp = 0;
        normalizePosition(ship);
        // FIXME: gravity wells?
    } else if (dist2 && ship.warp) {
        /* Normal movement */
        // First, compute new position in mx,my
        double dist = std::sqrt(double(dist2));
        int way = ship.warp * ship.warp;
        if (real_ship.canDoSpecial(hf_Gravitonic))
            way *= 2;

        int mx = ship.waypoint_dx, my = ship.waypoint_dy;
        if (dist > way) {
            int dx, dy;
            if (host.isTHost()) {
                // THost movement formulas, from Donovan's
                if (std::abs(mx) > std::abs(my)) {
                    dx = int(double(way) * std::abs(mx) / dist + 0.5);
                    dy = int(double(dx) * std::abs(my) / std::abs(mx) + 0.5);
                } else {
                    dy = int(double(way) * std::abs(my) / dist + 0.5);
                    dx = int(double(dy) * std::abs(mx) / std::abs(my) + 0.5);
                }
                if (mx < 0)
                    dx = -dx;
                if (my < 0)
                    dy = -dy;
            } else {
                // PHost. From docs and source.
                double head = getHeadingRad(mx, my);
                double fx = std::sin(head) * way;
                double fy = std::cos(head) * way;
                dx = int(fx);
                dy = int(fy);
                if (dx != mx)
                    dx += sgn(fx);
                if (dy != my)
                    dy += sgn(fy);
                if (mx == 0)
                    dx = 0;
                if (my == 0)
                    dy = 0;
            }

            // we now have the dx,dy we want to move
            mx = dx;
            my = dy;
            dist = way;         // FIXME: use distFromDX(dx,dy) instead?
        }

        /* Here, we have the position offset in mx,my. Compute fuel
           usage. We first have to move our towee, if any. */
        if (towee_override) {
            if (ship.mission == GMission::msn_Tow) {
                // we assume the tow succeeds. FIXME: be more clever?
                if (towee_override->ship.mission == GMission::msn_Tow)
                    towee_override->ship.mission = 0;
                towee_override->ship.warp = 0;
            }
            towee_override->computeTurn();
        }

        int fuel = computeFuelUsage(univ, ship,
                                    towee_override ? &towee_override->ship : 0,
                                    real_ship.canDoSpecial(hf_Gravitonic),
                                    dist);
        ship.ore_cargo[el_Neutronium] -= fuel;
        fuel_used += fuel;

        // okay, move it.
        ship.x += mx;
        ship.y += my;
        ship.waypoint_dx -= mx;
        ship.waypoint_dy -= my;
        normalizePosition(ship);

        // warp wells suck. (the ship to the planet, of course.)
        int wp_x = ship.x + ship.waypoint_dx;
        int wp_y = ship.y + ship.waypoint_dy;

        if (ship.warp > 1 && !univ.getPlanetAt(GPoint(ship.x, ship.y))) {
            if (int gpid = univ.getGravityPlanetAt(GPoint(ship.x, ship.y))) {
                // okay, there is a planet. Move the ship.
                GPoint ppos = univ.getPlanet(gpid).getPos();
                ship.x = ppos.x;
                ship.y = ppos.y;
                // Now adjust the waypoint. If the waypoint was inside
                // the warp well of the same planet, assume the end of
                // this movement order. Otherwise, when users set a
                // waypoint at the edge of the warp well, the ship
                // would try to get there for ever. This is consistent
                // with the CCScript `MoveTo' command and with what
                // most people expect.
                if (univ.getPlanetAt(GPoint(wp_x, wp_y), true) == gpid) {
                    ship.waypoint_dx = ship.waypoint_dy = 0;
                } else {
                    GPoint new_wp = GPoint(wp_x, wp_y).getNearestTo(GPoint(ship.x, ship.y));
                    ship.waypoint_dx = new_wp.x - ship.x;
                    ship.waypoint_dy = new_wp.y - ship.y;
                }
            }
        }
    } else {
        /* No sensible movement order for this ship. Advance towee's
           time anyway. */
        if (towee_override)
            towee_override->computeTurn();
    }

    /* turn is over. */
    ++turns_computed;
}

/** Compute this ship's movement, until movement is over or time runs
    out. */
void
GShipTurnPredictor::computeMovement()
{
    const int final_turn = turns_computed + MOVEMENT_TIME_LIMIT;
    while ((ship.waypoint_dx || ship.waypoint_dy) && turns_computed < final_turn) {
        computeTurn();
        if (ship.ore_cargo[el_Neutronium] < 0)
            ship.ore_cargo[el_Neutronium] = 0;
    }
}
