/**
  *  \file game/ship.h
  *  \brief Ships
  *
  *  This defines all ship-related classes.
  *
  *  (c) 2001-2006 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SHIP_H
#define PCC_V2_GAME_SHIP_H

#include "game/struct.h"
#include "game/obj.h"
#include "game/objtypes.h"
#include "game/spec.h"
#include "game/cargo.h"
#include "cpluslib/maybe.h"

class GUniverse;
class GTurnfile;
class GPlanet;

class GShip : public GMapObject {
 private:
    GShip();
    GShip(const GShip&);
    GShip& operator=(const GShip&);
 public:
    /// Transporters.
    enum Transporter {
        t_Unload,
        t_Transfer
    };
 private:
    int id;
    TShip* dat;
    TShipTarget* target;
    TShipXY shipxy;
    int orbit;
    /** Source flags. Bitfield (1<<player) of who's .dat file the
        information is from. Used for saving. */
    GPlayerSet src_flags;
    /** Target flags. Bitfield of who's TARGET file the information is
        from. */
    GPlayerSet target_flags;
    /** Remote control. 0 if not RCd, -1 if forbidden, otherwise real owner */
    int rc_flag;
    /** Cargo arbiter. */
    GCargo cargo;
 public:
    GShip(int id);
    ~GShip();

    /* Initialisation */
    void addData(const TShip& adat, GPlayerSet src_flags);
    void addTargetEntry(const TShipTarget& target, GPlayerSet flags);
    void addShipXYEntry(TShipXY axy);
    void addRCEntry(int flag);

    void getRawData(TShip& the_data) const
        { the_data = *dat; }

    /* Enquiry */
    bool haveData() const
        { return dat; }
    bool haveTarget() const
        { return target; }
    GPlayerSet getSrcFlags() const
        { return src_flags; }
    GPlayerSet getTargetFlags() const
        { return target_flags; }

    void applyTurnCommand(GTurnfile& trn, int index);
    void makeTurnCommands(GTurnfile& trn, const GShip& dis) const;

    /* Checking */
    void internalCheck();
    void combinedCheck1(GUniverse& univ);

    /* Accessors */
    int getMass() const;
    int getHullId() const;
    GHull& getHull() const
        { return ::getHull(getHullId()); }
    int getOrbit() const
        { return orbit; }
    int getRealOwner() const;

    GPoint getWaypoint() const;
    int getWaypointDX() const
        { return dat->waypoint_dx; }
    int getWaypointDY() const
        { return dat->waypoint_dy; }

    bool canDoSpecial(int hf) const;
    int getWarp() const;
    bool isHyperdriving() const;

    mp16_t getEngine() const
        { return dat ? mp16_t(dat->engine) : mp16_t(); }
    mp16_t getBeamType() const
        { return dat ? mp16_t(dat->beam_type) : mp16_t(); }
    mp16_t getBeamCount() const
        { return dat ? mp16_t(dat->beam_count) : mp16_t(); }
    mp16_t getBayCount() const
        { return dat ? mp16_t(dat->bay_count) : mp16_t(); }
    mp16_t getTorpType() const
        { return dat ? mp16_t(dat->torp_type) : mp16_t(); }
    mp16_t getTorpLauncherCount()
        const { return dat ? mp16_t(dat->torp_launcher_count) : mp16_t(); }
    int getAmmo() const; // { return dat->ammo; }

    mp16_t getMission() const
        { return dat ? mp16_t(dat->mission) : mp16_t(); }
    mp16_t getTowId() const
        { return dat ? mp16_t(dat->mission_tow) : mp16_t(); }
    mp16_t getIntId() const
        { return dat ? mp16_t(dat->mission_intercept) : mp16_t(); }
    mp16_t getPE() const
        { return dat ? mp16_t(dat->primary_enemy) : mp16_t(); }
    int    getDamage() const { return dat->damage; }
    int    getCrew() const { return dat->crew; }
    int    getColonists() const { return getCargo(el_Colonists); }
    int    getSupplies() const { return getCargo(el_Supplies); }
    int    getMoney() const { return getCargo(el_Money); }
    int    getCargoRaw(GCargoType type) const;
    int    getCargo(GCargoType type) const;

    int    getFreeCargo() const;

    string_t getFCode() const
        { return getBasicString(dat->fcode); }

    GCargo& getCargoHandler()
        { return cargo; }

    TShipTransfer& getTransporter(Transporter which_one) throw()
        { return which_one == t_Transfer ? dat->transfer : dat->unload; }
    const TShipTransfer& getTransporter(Transporter which_one) const throw()
        { return which_one == t_Transfer ? dat->transfer : dat->unload; }
    bool isTransporterActive(Transporter which_one) const throw();
    bool isTransporterCancellable(Transporter which_one) const throw();
    void cancelTranporter(Transporter which_one) throw();

    bool   getWaypointPermission() const;

    bool   isCloningAt(const GPlanet& planet) const;

    /* Setters */
    void   setName(const string_t& str);
    void   setFCode(const string_t& str);
    void   setPE(int pe);
    void   setWaypoint(GPoint pt);
    void   setWarp(int warp);
    void   changeCargoRaw(GCargoType type, int amount);
    void   setMission(int m, int i, int t);

    void   leaveFleet();

    /* GMapObject interface */
    GPoint   getPos() const;
    string_t getName() const;
    string_t getFullName() const;
    int      getId() const;
    int      getOwner() const;
    bool     isExistant() const;

    /** Ship flags. */
    enum {
        ss_DirtyData = 1,
        ss_DirtyHead = 2,
        ss_DirtyXY = 4
    };

    /** Ship facilities. */
    enum Facility {
        sf_Unload,              ///< Unload transporter.
        sf_Transfer             ///< Ship/Ship transporter.
    };
};

class GShipType : public GObjectType {
 public:
    GShipType(GUniverse& univ);

    bool isExistant(int index);
    string_t getClassName();
    GShip& getObject(int index);
    GShip& getShip()
        { return getObject(getCurrent()); }
};

class GShipTransfer : public GCargoContainer {
    GShip& ship;
 protected:
    int32  getAmount(GCargoType type);
    void   changeAmount(GCargoType type, int32 amount);
    void   reserveAmount(GCargoType type, int32 amount);
 public:
    GShipTransfer(GShip& ship);
    ~GShipTransfer();

    bool     canHave(GCargoType cargo_id);
    int32    getMax(GCargoType cargo_id);
    string_t getName() const;
};

class GShipTransporter : public GCargoContainer {
    GShip&             ship;     //!< The ship whose transporter we're using.
    TShipTransfer&     transfer; //!< The transporter we're using
    GObject*           other;    //!< The other object (null if jettison)
    GShip::Transporter which_one;
    TShipTransfer      reserve;
 protected:
    int32  getAmount(GCargoType type);
    void   changeAmount(GCargoType type, int32 amount);
    void   reserveAmount(GCargoType type, int32 amount);
 public:
    GShipTransporter(GShip& ship, GShip::Transporter which_one, GObject& other);
    GShipTransporter(GShip& ship);
    ~GShipTransporter();

    bool     canHave(GCargoType cargo_id);
    int32    getMax(GCargoType cargo_id);
    string_t getName() const;
};

class GShipTurnPredictor {
    int                 id;
    TShip               ship;
    GShipTurnPredictor* towee_override;
    GUniverse&          univ;
    int32               fuel_used;
    int                 turns_computed;

 public:
    GShipTurnPredictor(GUniverse& turn, int id);
    GShipTurnPredictor(GUniverse& turn, int id, GShipTurnPredictor& towee_override);

    /** Get total fuel used during computation. */
    int32 getFuelUsed() const
        { return fuel_used; }

    /** Get number of turns computed. */
    int getTurnCount() const
        { return turns_computed; }

    void  computeTurn();
    void  computeMovement();

    static const int MOVEMENT_TIME_LIMIT = 30;
};

#endif
